==> hlog-0.20-rc1.tar.gz: hlog-0.20-rc1/scripts/export-pota-adif.lua <==
--
-- Copyright (c) 2020-2021 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
-- SOFTWARE.
--

local TMPDIR = "/tmp"
local HIST_HEIGHT = 10

local us_states = {
	"AK", "AL", "AR", "AZ",
	"CA", "CO", "CT",
	"DC", "DE",
	"FL",
	"GA",
	"HI",
	"IA", "ID", "IL", "IN",
	"KS", "KY",
	"LA",
	"MA", "MD", "ME", "MI", "MN", "MO", "MS", "MT",
	"NC", "ND", "NE", "NH", "NJ", "NM", "NV", "NY",
	"OH", "OK", "OR",
	"PA",
	"RI",
	"SC", "SD",
	"TN", "TX",
	"UT",
	"VA", "VT",
	"WA", "WI", "WV", "WY",
}

local ca_provinces = {
	-- actual provinces
	"AB",
	"BC",
	"MB",
	"NB", "NL", "NS",
	"ON",
	"PE",
	"QC",
	"SK",

	-- territories
	"NT", "NU",
	"YT",
}

local function max(a, b)
	return (a > b) and a or b
end

local function padnum(v, width)
	local tmp = tostring(v)

	while tmp:len() < width do
		tmp = " " .. tmp
	end

	return tmp
end

local function padnum_max(c, m)
	c = tostring(c)
	m = tostring(m)

	if c:len() < m:len() then
		return string.rep(" ", m:len() - c:len()) .. c
	end

	return c
end

local function tcount(t)
	local n = 0

	if t == nil then
		return 0
	end

	for _, _ in pairs(t) do
		n = n + 1
	end

	return n
end

local function sorted_keys(t)
	local o = {}

	for k, _ in pairs(t) do
		table.insert(o, k)
	end

	table.sort(o)

	return o
end

local function bandname(b)
	local names = {
		[80000] = "80m",
		[40000] = "40m",
		[20000] = "20m",
		[15000] = "15m",
		[10000] = "10m",
	}

	return (names[b] ~= nil) and names[b] or string.format("%s", b / 1000)
end

-- Generate a unique key for a contact.
local function gen_key(qso)
	local call = qso.rx.station_call
	local band = qso.tx.band / 1000
	local mode = qso.tx.mode
	local rxref = qso.additional['sig_info']

	return string.format("%s-%s-%s-%s", call, band, mode,
			     (rxref ~= nil) and rxref or "NONE")
end

-- Drop the various "mobile" and "portable" suffixes.  Not only does POTA
-- not require them, it gets a little confused by them.  Specifically, it'll
-- break the activity/activation into one per station call.  This can be
-- worked around in the POTA UI by adding additional calls to your account,
-- but that's annoying.  So, we pretend the call was without one of these
-- common and trivial suffixes.
local function simplify_station_call(call)
	for _, suffix in ipairs({ "P", "M", "MM", "AM" }) do
		if call:match("/"..suffix.."$") ~= nil then
			return call:sub(1, -2 - suffix:len())
		end
	end

	return call
end

-- Sometimes one can be in more than one reference area at the same time.
-- Split the input string (comma separated refs) into an array.
local function get_refs(input)
	if input == nil then
		return nil
	end

	local out = {}

	while true do
		local comma = input:find(",", 1, true)
		if comma == nil then
			break
		end

		table.insert(out, input:sub(1, comma - 1))
		input = input:sub(comma + 1)
	end

	table.insert(out, input)

	return out
end

local function do_add_entry(prog, signame, station, orig_qso, txref, rxref)
	local basename = string.format("%s_%s_%s_%04u%02u%02u",
				       orig_qso.tx.operator_call,
				       txref,
				       orig_qso.tx.state,
				       orig_qso.start_time.year,
				       orig_qso.start_time.month,
				       orig_qso.start_time.day)

	-- get a writable copy
	qso = orig_qso:clone()
	hlog.nvl.clear(qso.additional)
	if txref ~= nil then
		qso.additional['my_sig'] = signame
		qso.additional['my_sig_info'] = txref
	end
	if rxref ~= nil then
		qso.additional['sig'] = signame
		qso.additional['sig_info'] = rxref
	end

	-- drop /P, /M, etc. from station call
	qso.tx.station_call = station

	-- clear undesired qso & side fields
	qso.comment = nil

	if qso.rx.freq == qso.tx.freq then
		qso.rx.freq = nil
	end
	if qso.rx.mode == qso.tx.mode then
		qso.rx.mode = nil
	end

	local cur = prog.logs[basename]
	if cur == nil then
		cur = {
			chrono = {}, -- array of QSOs, chronologically ordered
			dedup = {}, -- <key,count> dedup counter
			states = {}, -- <state,count> rx states stats
			dxcc = {}, -- <dxcc,count> rx dxcc stats
			p2p = {}, -- <call,rx ref> park-to-park stats
			bandmode = {}, -- <band,<mode,count>>
		}
		prog.logs[basename] = cur
	end

	-- We don't have to include date & tx ref, since those are already
	-- taken care of.
	local key = gen_key(qso)

	-- add QSO to chronological array
	table.insert(cur.chrono, qso)
	prog.mangled_qso_to_orig[qso] = orig_qso

	-- account for QSO in dedup map
	cur.dedup[key] = (cur.dedup[key] ~= nil) and (cur.dedup[key] + 1) or 1

	-- account for QSO dxcc
	if qso.rx.dxcc ~= nil then
		local dxcc = string.format("%s (%s)", qso.rx.country, qso.rx.dxcc)

		cur.dxcc[dxcc] = (cur.dxcc[dxcc] ~= nil) and (cur.dxcc[dxcc] + 1) or 1
	end

	-- account for QSO state
	if qso.rx.state ~= nil then
		local st = qso.rx.state

		cur.states[st] = (cur.states[st] ~= nil) and (cur.states[st] + 1) or 1
	end

	-- account for park-to-park stats
	if rxref ~= nil then
		if cur.p2p[qso.rx.station_call] == nil then
			cur.p2p[qso.rx.station_call] = {}
		end

		table.insert(cur.p2p[qso.rx.station_call], rxref)
	end

	-- account for band & mode
	local mode = qso.tx.mode
	local band = bandname(qso.tx.band)

	if cur.bandmode[band] == nil then
		cur.bandmode[band] = {}
	end
	if cur.bandmode[band][mode] == nil then
		cur.bandmode[band][mode] = 0
	end
	cur.bandmode[band][mode] = cur.bandmode[band][mode] + 1

	--
	-- Finally, clean up the QSO struct, *after* we did our stats
	-- accounting.
	--
	for _, key in ipairs({ "name", "power", "qth", "country", "county",
			       "continent", "dxcc", "itu", "cqz", "seq" }) do
		qso.tx[key] = nil
		qso.rx[key] = nil
	end
end

local function add_entry(qso, prog, signame, tx, rx, submit, allow_multiple)
	-- needs to be a contact that we haven't exported before or are
	-- labeled as bad
	if (qso.additional[tx] == nil) or
	   (qso.additional[submit] ~= nil) or
	   (qso.additional['bad'] ~= nil) then
		return
	end

	assert(qso.tx.station_call ~= nil)

	local station = simplify_station_call(qso.tx.station_call)

	-- TODO: check for dups

	local txrefs = get_refs(qso.additional[tx])
	local rxrefs = get_refs(qso.additional[rx])

	assert(txrefs ~= nil)

	if not allow_multiple and tcount(txrefs) > 1 then
		print(string.format("Warning: %s QSO with multiple tx WWFF references (%s)",
				    qso.uuid, qso.additional['tx_wwff']))
	end
	if not allow_multiple and tcount(rxrefs) > 1 then
		print(string.format("Warning: %s QSO multiple rx WWFF references",
				    qso.uuid, qso.additional['rx_wwff']))
	end

	for _, txref in ipairs(txrefs) do
		if rxrefs == nil then
			-- regular hunter
			do_add_entry(prog, signame, station, qso, txref, nil)
		else
			-- P2P contact, handle the other activator
			-- being in multiple entities concurrently
			for _, rxref in ipairs(rxrefs) do
				do_add_entry(prog, signame, station, qso, txref, rxref)
			end
		end
	end
end

local function write_dxcc_stats(f, dxccs)
	f:write("DXCC stats:\n")
	f:write("  Count Country (DXCC)\n")
	local by_count = {}
	for dxcc, count in pairs(dxccs) do
		local tmp = by_count[count]
		if tmp == nil then
			by_count[count] = {}
		end
		table.insert(by_count[count], dxcc)
	end

	for _, count in ipairs(sorted_keys(by_count)) do
		local dxccs = by_count[count]
		table.sort(dxccs)

		for _, dxcc in ipairs(dxccs) do
			f:write(string.format("  %s %s\n", padnum(count, 5), dxcc))
		end
	end
	f:write("\n")
end

local function write_state_stats(f, states)
	local function count_states(allstates)
		local count = 0
		local maxstate = 0
		for _, state in ipairs(allstates) do
			if states[state] then
				count = count + 1
				maxstate = math.max(maxstate, states[state])
			end
		end

		return count, maxstate
	end

	local function collate_states(allstates, maxstate)
		local got = {}
		local mis = {}
		local hist = {}
		for _, state in ipairs(allstates) do
			if states[state] ~= nil then
				local hashes = math.ceil(HIST_HEIGHT * states[state] / maxstate)
				local blanks = HIST_HEIGHT - hashes

				table.insert(got, string.format("%s(%u)", state, states[state]))
				table.insert(hist,
					     string.rep(" ", blanks) ..
					     string.rep("#", hashes) ..
					     "-" ..
					     state ..
					     padnum_max(states[state], maxstate))
			else
				table.insert(mis, string.format("%s", state))
				table.insert(hist,
					     string.rep(" ", HIST_HEIGHT) ..
					     "-" ..
					     state ..
					     string.rep(" ", tostring(maxstate):len() - 1).."-")
			end
		end

		return got, mis, hist
	end

	local function list_states(pfx, list)
		local out = "  " .. pfx
		for idx, state in ipairs(list) do
			out = out .. " " .. state

			if out:len() > 60 then
				f:write(out .. "\n")
				out = "       "
			end
		end
		f:write(out .. "\n")
	end

	local function print_summary(label, count, allstates, got, mis)
		f:write(string.format("%s stats: %2u/%2u\n", label, count, tcount(allstates)))

		list_states("Got: ", got)
		list_states("Need:", mis)
	end

	-- find number of states and most frequent state
	local us_count, maxstate1 = count_states(us_states)
	local ca_count, maxstate2 = count_states(ca_provinces)
	local maxstate = max(maxstate1, maxstate2)

	if count == 0 then
		return -- don't print state stats if we have none
	end

	-- output state counts
	us_got, us_mis, us_hist = collate_states(us_states, maxstate)
	ca_got, ca_mis, ca_hist = collate_states(ca_provinces, maxstate)

	print_summary("State", us_count, us_states, us_got, us_mis)
	f:write("\n")
	print_summary("Province/territory", ca_count, ca_provinces, ca_got, ca_mis)

	-- histogram
	f:write(string.format("\n  Each '#' = %.1f contacts:\n", maxstate / HIST_HEIGHT))
	for row = 1, HIST_HEIGHT + 1 + 2 + tostring(maxstate):len() do
		local out = "  "
		for _, h in ipairs(us_hist) do
			out = out .. h:sub(row, row)
		end
		out = out .. " "
		for _, h in ipairs(ca_hist) do
			out = out .. h:sub(row, row)
		end
		f:write(out .. "\n")
	end
end

local function write_bandmode_stats(f, cur)
	-- get a sorted list of all subkeys
	local allb = {}
	local widths = {}
	for a, _ in pairs(cur) do
		for b, c in pairs(cur[a]) do
			allb[b] = true
			widths[b] = max(b:len(), tostring(c):len())
		end
	end
	allb = sorted_keys(allb)

	-- print header
	local out = string.format("  %4s ", "")
	for _, b in ipairs(allb) do
		out = out .. string.format("%s ", b)
	end
	f:write(out .. "| Total\n")

	-- print data rows
	for _, a in ipairs(sorted_keys(cur)) do
		local total = 0
		local out = string.format("  %-4s ", a)
		for _, b in ipairs(allb) do
			if cur[a][b] ~= nil then
				total = total + cur[a][b]
				out = out .. string.format("%s ", padnum(cur[a][b], widths[b]))
			else
				out = out .. string.format("%s ", padnum("", widths[b]))
			end
		end

		f:write(string.format("%s| %s\n", out, padnum(total, 5)))
	end
end

local function open_uuid_file(f, basename, suffix)
	if f ~= nil then
		return f
	end

	return assert(io.open(basename .. suffix, "w"))
end

local function close_uuid_file(f)
	if f ~= nil then
		f:close()
	end
end

local function gen_logs(prog, progname)
	local logs = prog.logs

	for basename,cur in pairs(logs) do
		print(basename)

		-- generate files with all the uuids submitted
		local logpath = TMPDIR .. "/" .. basename
		local uuids = nil
		local f = assert(io.open(logpath .. ".adif", "w"))

		-- Statistics
		f:write(string.format("%s ADIF export - %s\n",
				      progname, basename))

		-- summary
		f:write(string.format("%u contacts, %u unique, %u dups\n",
				      tcount(cur.chrono), tcount(cur.dedup),
				      tcount(cur.chrono) - tcount(cur.dedup)))

		-- duplicates
		for key, count in pairs(cur.dedup) do
			assert(count ~= 0)

			if count > 1 then
				f:write(string.format("%u dups for %s\n", count - 1, key))
			end
		end
		f:write("\n")

		-- bands & modes
		f:write("Bands & modes:\n")
		write_bandmode_stats(f, cur.bandmode)
		f:write("\n")

		-- park-to-park stats
		if tcount(cur.p2p) > 0 then
			f:write("Park-to-park:\n")

			for _, call in ipairs(sorted_keys(cur.p2p)) do
				local refs = cur.p2p[call]
				table.sort(refs)

				local out = string.format("  %s (%d) @",
							  call, tcount(refs))
				for _, ref in ipairs(refs) do
					out = out .. " " .. ref
				end
				f:write(out .. "\n")
			end
			f:write("\n")
		end

		-- dxcc stats
		write_dxcc_stats(f, cur.dxcc)

		-- state stats
		write_state_stats(f, cur.states)

		-- ADIF header
		f:write(hlog.adif.write_header())
		f:write("\n")

		-- ADIF QSOs
		for _, qso in pairs(cur.chrono) do
			-- note the QSO uuid
			uuids = open_uuid_file(uuids, logpath, ".uuids")

			uuids:write(qso.uuid .. "\n")

			-- Do *not* use strict mode, POTA doesn't check the submode field
			-- and the strict mode field doesn't include enough info (e.g.,
			-- PSK31 and QPSK31 both use "PSK" as the mode).
			f:write(hlog.adif.write_qso(qso, false))
			f:write("\n")
		end

		f:close()
		close_uuid_file(uuids)
	end
end

function main()
	local pota = {
		logs = {},
		mangled_qso_to_orig = {},
	}
	local wwff = {
		logs = {},
		mangled_qso_to_orig = {},
	}

	for qso in hlog.index.history(true) do
		add_entry(qso, pota, 'POTA', 'tx_pota', 'rx_pota', 'pota_log_submitted', true)
		add_entry(qso, wwff, 'WWFF', 'tx_wwff', 'rx_wwff', 'wwff_log_submitted', false)
	end

	gen_logs(pota, 'POTA')
	gen_logs(wwff, 'WWFF')
end

==> hlog-0.21-rc1.tar.gz: hlog-0.21-rc1/scripts/export-pota-adif.lua <==
--
-- Copyright (c) 2020-2022 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
-- SOFTWARE.
--

local TMPDIR = "/tmp"
local HIST_HEIGHT = 10

local us_states = {
	"AK", "AL", "AR", "AZ",
	"CA", "CO", "CT",
	"DC", "DE",
	"FL",
	"GA",
	"HI",
	"IA", "ID", "IL", "IN",
	"KS", "KY",
	"LA",
	"MA", "MD", "ME", "MI", "MN", "MO", "MS", "MT",
	"NC", "ND", "NE", "NH", "NJ", "NM", "NV", "NY",
	"OH", "OK", "OR",
	"PA",
	"RI",
	"SC", "SD",
	"TN", "TX",
	"UT",
	"VA", "VT",
	"WA", "WI", "WV", "WY",
}

local ca_provinces = {
	-- actual provinces
	"AB",
	"BC",
	"MB",
	"NB", "NL", "NS",
	"ON",
	"PE",
	"QC",
	"SK",

	-- territories
	"NT", "NU",
	"YT",
}

local function max(a, b)
	return (a > b) and a or b
end

local function padnum(v, width)
	local tmp = tostring(v)

	while tmp:len() < width do
		tmp = " " .. tmp
	end

	return tmp
end

local function padnum_max(c, m)
	c = tostring(c)
	m = tostring(m)

	if c:len() < m:len() then
		return string.rep(" ", m:len() - c:len()) .. c
	end

	return c
end

local function tcount(t)
	local n = 0

	if t == nil then
		return 0
	end

	for _, _ in pairs(t) do
		n = n + 1
	end

	return n
end

local function sorted_keys(t)
	local o = {}

	for k, _ in pairs(t) do
		table.insert(o, k)
	end

	table.sort(o)

	return o
end

-- Generate a unique key for a contact.
local function gen_key(qso)
	local call = qso.rx.station_call
	local band = qso.tx.band / 1000
	local mode = qso.tx.mode
	local rxref = qso.additional['sig_info']

	return string.format("%s-%s-%s-%s", call, band, mode,
			     (rxref ~= nil) and rxref or "NONE")
end

-- Drop the various "mobile" and "portable" suffixes.  Not only does POTA
-- not require them, it gets a little confused by them.  Specifically, it'll
-- break the activity/activation into one per station call.  This can be
-- worked around in the POTA UI by adding additional calls to your account,
-- but that's annoying.  So, we pretend the call was without one of these
-- common and trivial suffixes.
local function simplify_station_call(call)
	for _, suffix in ipairs({ "P", "M", "MM", "AM" }) do
		if call:match("/"..suffix.."$") ~= nil then
			return call:sub(1, -2 - suffix:len())
		end
	end

	return call
end

-- Sometimes one can be in more than one reference area at the same time.
-- Split the input string (comma separated refs) into an array.
local function get_refs(input)
	if input == nil then
		return nil
	end

	local out = {}

	while true do
		local comma = input:find(",", 1, true)
		if comma == nil then
			break
		end

		table.insert(out, input:sub(1, comma - 1))
		input = input:sub(comma + 1)
	end

	table.insert(out, input)

	return out
end

local function do_add_entry(prog, signame, station, orig_qso, txref, rxref)
	local basename = string.format("%s_%s_%s_%04u%02u%02u",
				       orig_qso.tx.operator_call,
				       txref:gsub('/', '-'),
				       orig_qso.tx.state,
				       orig_qso.start_time.year,
				       orig_qso.start_time.month,
				       orig_qso.start_time.day)

	-- get a writable copy
	qso = orig_qso:clone()
	hlog.nvl.clear(qso.additional)
	if signame == "SOTA" then
		qso.additional['my_sota_ref'] = txref
		qso.additional['sota_ref'] = rxref
	else
		if txref ~= nil then
			qso.additional['my_sig'] = signame
			qso.additional['my_sig_info'] = txref
		end
		if rxref ~= nil then
			qso.additional['sig'] = signame
			qso.additional['sig_info'] = rxref
		end
	end

	-- drop /P, /M, etc. from station call
	qso.tx.station_call = station

	-- clear undesired qso & side fields
	qso.comment = nil

	if qso.rx.freq == qso.tx.freq then
		qso.rx.freq = nil
	end
	if qso.rx.mode == qso.tx.mode then
		qso.rx.mode = nil
	end

	local cur = prog.logs[basename]
	if cur == nil then
		cur = {
			chrono = {}, -- array of QSOs, chronologically ordered
			dedup = {}, -- <key,count> dedup counter
			states = {}, -- <state,count> rx states stats
			dxcc = {}, -- <dxcc,count> rx dxcc stats
			p2p = {}, -- <call,rx ref> park-to-park stats
			bandmode = {}, -- <band,<mode,count>>
		}
		prog.logs[basename] = cur
	end

	-- We don't have to include date & tx ref, since those are already
	-- taken care of.
	local key = gen_key(qso)

	-- add QSO to chronological array
	table.insert(cur.chrono, qso)
	prog.mangled_qso_to_orig[qso] = orig_qso

	-- account for QSO in dedup map
	cur.dedup[key] = (cur.dedup[key] ~= nil) and (cur.dedup[key] + 1) or 1

	-- account for QSO dxcc
	if qso.rx.dxcc ~= nil then
		local dxcc = string.format("%s (%s)", qso.rx.country, qso.rx.dxcc)

		cur.dxcc[dxcc] = (cur.dxcc[dxcc] ~= nil) and (cur.dxcc[dxcc] + 1) or 1
	end

	-- account for QSO state
	if qso.rx.state ~= nil then
		local st = qso.rx.state

		cur.states[st] = (cur.states[st] ~= nil) and (cur.states[st] + 1) or 1
	end

	-- account for park-to-park stats
	if rxref ~= nil then
		if cur.p2p[qso.rx.station_call] == nil then
			cur.p2p[qso.rx.station_call] = {}
		end

		table.insert(cur.p2p[qso.rx.station_call], rxref)
	end

	-- account for band & mode
	local mode = qso.tx.mode
	local band = qso.tx.band_name

	if cur.bandmode[band] == nil then
		cur.bandmode[band] = {}
	end
	if cur.bandmode[band][mode] == nil then
		cur.bandmode[band][mode] = 0
	end
	cur.bandmode[band][mode] = cur.bandmode[band][mode] + 1

	--
	-- Finally, clean up the QSO struct, *after* we did our stats
	-- accounting.
	--
	for _, key in ipairs({ "name", "power", "qth", "country", "county",
			       "continent", "dxcc", "itu", "cqz", "seq" }) do
		qso.tx[key] = nil
		qso.rx[key] = nil
	end
end

local function add_entry(qso, prog, signame, tx, rx, submit, allow_multiple)
	-- needs to be a contact that we haven't exported before or are
	-- labeled as bad
	if (qso.additional[tx] == nil) or
	   (qso.additional[submit] ~= nil) or
	   (qso.additional['bad'] ~= nil) then
		return
	end

	if qso.tx.band == nil then
		print(string.format("Error: %s QSO lacks band info", qso.uuid))
		return
	end

	assert(qso.tx.station_call ~= nil)

	local station = simplify_station_call(qso.tx.station_call)

	-- TODO: check for dups

	local txrefs = get_refs(qso.additional[tx])
	local rxrefs = get_refs(qso.additional[rx])

	assert(txrefs ~= nil)

	if not allow_multiple and tcount(txrefs) > 1 then
		print(string.format("Warning: %s QSO with multiple tx %s references (%s)",
				    qso.uuid, signame, qso.additional[tx]))
	end
	if not allow_multiple and tcount(rxrefs) > 1 then
		print(string.format("Warning: %s QSO multiple rx %s references (%s)",
				    qso.uuid, signame, qso.additional[rx]))
	end

	for _, txref in ipairs(txrefs) do
		if rxrefs == nil then
			-- regular hunter
			do_add_entry(prog, signame, station, qso, txref, nil)
		else
			-- P2P contact, handle the other activator
			-- being in multiple entities concurrently
			for _, rxref in ipairs(rxrefs) do
				do_add_entry(prog, signame, station, qso, txref, rxref)
			end
		end
	end
end

local function write_dxcc_stats(f, dxccs)
	f:write("DXCC stats:\n")
	f:write("  Count Country (DXCC)\n")
	local by_count = {}
	for dxcc, count in pairs(dxccs) do
		local tmp = by_count[count]
		if tmp == nil then
			by_count[count] = {}
		end
		table.insert(by_count[count], dxcc)
	end

	for _, count in ipairs(sorted_keys(by_count)) do
		local dxccs = by_count[count]
		table.sort(dxccs)

		for _, dxcc in ipairs(dxccs) do
			f:write(string.format("  %s %s\n", padnum(count, 5), dxcc))
		end
	end
	f:write("\n")
end

local function write_state_stats(f, states)
	local function count_states(allstates)
		local count = 0
		local maxstate = 0
		for _, state in ipairs(allstates) do
			if states[state] then
				count = count + 1
				maxstate = math.max(maxstate, states[state])
			end
		end

		return count, maxstate
	end

	local function collate_states(allstates, maxstate)
		local got = {}
		local mis = {}
		local hist = {}
		for _, state in ipairs(allstates) do
			if states[state] ~= nil then
				local hashes = math.ceil(HIST_HEIGHT * states[state] / maxstate)
				local blanks = HIST_HEIGHT - hashes

				table.insert(got, string.format("%s(%u)", state, states[state]))
				table.insert(hist,
					     string.rep(" ", blanks) ..
					     string.rep("#", hashes) ..
					     "-" ..
					     state ..
					     padnum_max(states[state], maxstate))
			else
				table.insert(mis, string.format("%s", state))
				table.insert(hist,
					     string.rep(" ", HIST_HEIGHT) ..
					     "-" ..
					     state ..
					     string.rep(" ", tostring(maxstate):len() - 1).."-")
			end
		end

		return got, mis, hist
	end

	local function list_states(pfx, list)
		local out = "  " .. pfx
		local neednl = true

		for idx, state in ipairs(list) do
			out = out .. " " .. state
			neednl = true

			if out:len() > 60 then
				f:write(out .. "\n")
				out = "       "
				neednl = false
			end
		end

		if neednl then
			f:write(out .. "\n")
		end
	end

	local function print_summary(label, count, allstates, got, mis)
		f:write(string.format("%s stats: %2u/%2u\n", label, count, tcount(allstates)))

		list_states("Got: ", got)
		list_states("Need:", mis)
	end

	-- find number of states and most frequent state
	local us_count, maxstate1 = count_states(us_states)
	local ca_count, maxstate2 = count_states(ca_provinces)
	local maxstate = max(maxstate1, maxstate2)

	if count == 0 then
		return -- don't print state stats if we have none
	end

	-- output state counts
	us_got, us_mis, us_hist = collate_states(us_states, maxstate)
	ca_got, ca_mis, ca_hist = collate_states(ca_provinces, maxstate)

	print_summary("State", us_count, us_states, us_got, us_mis)
	f:write("\n")
	print_summary("Province/territory", ca_count, ca_provinces, ca_got, ca_mis)

	-- histogram
	f:write(string.format("\n  Each '#' = %.1f contacts:\n", maxstate / HIST_HEIGHT))
	for row = 1, HIST_HEIGHT + 1 + 2 + tostring(maxstate):len() do
		local out = "  "
		for _, h in ipairs(us_hist) do
			out = out .. h:sub(row, row)
		end
		out = out .. " "
		for _, h in ipairs(ca_hist) do
			out = out .. h:sub(row, row)
		end
		f:write(out .. "\n")
	end
end

local function write_bandmode_stats(f, cur)
	-- get a sorted list of all subkeys
	local allb = {}
	local widths = {}
	for a, _ in pairs(cur) do
		for b, c in pairs(cur[a]) do
			allb[b] = true
			widths[b] = max(b:len(), tostring(c):len())
		end
	end
	allb = sorted_keys(allb)

	-- print header
	local out = string.format("  %4s ", "")
	for _, b in ipairs(allb) do
		out = out .. string.format("%s ", b)
	end
	f:write(out .. "| Total\n")

	-- print data rows
	for _, a in ipairs(sorted_keys(cur)) do
		local total = 0
		local out = string.format("  %-4s ", a)
		for _, b in ipairs(allb) do
			if cur[a][b] ~= nil then
				total = total + cur[a][b]
				out = out .. string.format("%s ", padnum(cur[a][b], widths[b]))
			else
				out = out .. string.format("%s ", padnum("", widths[b]))
			end
		end

		f:write(string.format("%s| %s\n", out, padnum(total, 5)))
	end
end

local function open_uuid_file(f, basename, suffix)
	if f ~= nil then
		return f
	end

	return assert(io.open(basename .. suffix, "w"))
end

local function close_uuid_file(f)
	if f ~= nil then
		f:close()
	end
end

local function gen_logs(prog, progname)
	local logs = prog.logs

	for basename,cur in pairs(logs) do
		print(basename)

		-- generate files with all the uuids submitted
		local logpath = TMPDIR .. "/" .. basename
		local uuids = nil
		local f = assert(io.open(logpath .. ".adif", "w"))

		-- Statistics
		f:write(string.format("%s ADIF export - %s\n",
				      progname, basename))

		-- summary
		f:write(string.format("%u contacts, %u unique, %u dups\n",
				      tcount(cur.chrono), tcount(cur.dedup),
				      tcount(cur.chrono) - tcount(cur.dedup)))

		-- duplicates
		for key, count in pairs(cur.dedup) do
			assert(count ~= 0)

			if count > 1 then
				f:write(string.format("%u dups for %s\n", count - 1, key))
			end
		end
		f:write("\n")

		-- bands & modes
		f:write("Bands & modes:\n")
		write_bandmode_stats(f, cur.bandmode)
		f:write("\n")

		-- park-to-park stats
		if tcount(cur.p2p) > 0 then
			f:write("Park-to-park:\n")

			for _, call in ipairs(sorted_keys(cur.p2p)) do
				local refs = cur.p2p[call]
				table.sort(refs)

				local out = string.format("  %s (%d) @",
							  call, tcount(refs))
				for _, ref in ipairs(refs) do
					out = out .. " " .. ref
				end
				f:write(out .. "\n")
			end
			f:write("\n")
		end

		-- dxcc stats
		write_dxcc_stats(f, cur.dxcc)

		-- state stats
		write_state_stats(f, cur.states)

		-- ADIF header
		f:write(hlog.adif.write_header())
		f:write("\n")

		-- ADIF QSOs
		for _, qso in pairs(cur.chrono) do
			-- note the QSO uuid
			uuids = open_uuid_file(uuids, logpath, ".uuids")

			uuids:write(qso.uuid .. "\n")

			-- Do *not* use strict mode, POTA doesn't check the submode field
			-- and the strict mode field doesn't include enough info (e.g.,
			-- PSK31 and QPSK31 both use "PSK" as the mode).
			f:write(hlog.adif.write_qso(qso, false))
			f:write("\n")
		end

		f:close()
		close_uuid_file(uuids)
	end
end

function main()
	local pota = {
		logs = {},
		mangled_qso_to_orig = {},
	}
	local wwff = {
		logs = {},
		mangled_qso_to_orig = {},
	}
	local sota = {
		logs = {},
		mangled_qso_to_orig = {},
	}

	for qso in hlog.index.history(true) do
		add_entry(qso, pota, 'POTA', 'tx_pota', 'rx_pota', 'pota_log_submitted', true)
		add_entry(qso, wwff, 'WWFF', 'tx_wwff', 'rx_wwff', 'wwff_log_submitted', false)
		add_entry(qso, sota, 'SOTA', 'tx_sota', 'rx_sota', 'sota_log_submitted', false)
	end

	gen_logs(pota, 'POTA')
	gen_logs(wwff, 'WWFF')
	gen_logs(sota, 'SOTA')
end
